import numpy as np
from collections.itertools import combinations
import matplotlib.pyplot as plt
from .calc_manager import LightingCalculator


class Zone:
    """Flags + shared behaviour common to ALL zones (planes, volumes, …)."""

    def __init__(
        self,
        zone_id=None,
        name=None,
        offset=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
        **kwargs,
    ):

        self.zone_id = str(zone_id)
        self.name = zone_id if name is None else name
        self.offset = True if offset is None else offset
        self.dose = False if dose is None else dose
        if self.dose:
            self.units = "mJ/cm²"
        else:
            self.units = "uW/cm²"
        self.hours = 8.0 if hours is None else abs(hours)  # only used if dose is true
        self.enabled = True if enabled is None else enabled
        self.show_values = True if show_values is None else show_values

        # may need a different lighting calculator defined for Planes and Volumes...
        self.calculator = LightingCalculator(self)

    def to_dict(self):

        data = {}
        data["zone_id"] = self.zone_id
        data["name"] = self.name
        data["offset"] = self.offset
        data["dose"] = self.dose
        data["hours"] = self.hours
        data["enabled"] = self.enabled
        data["show_values"] = self.show_values

        data.update(self._extra_dict())

    def _extra_dict(self):
        return {}


class Plane(Zone):
    def __init__(
        self,
        *,  # keyword only to avoid arg clashes
        vert: bool = False,
        horiz: bool = False,
        fov_vert: float = 180.0,
        fov_horiz: float = 360.0,
        height: float = 0.0,
        normal: np.ndarray | None = None,
        **zone_flags,  # anything Zone cares about
    ):
        # set generic flags first
        super().__init__(**zone_flags)

    def plot(self):
        """"""
        pass  # tmp

    def export(self, fname):
        """"""
        pass  # tmp

    def _extra_dict(self):

        zone_data = super()._extra_dict()

        data = {
            "fov_vert": self.fov_vert,
            "fov_horiz": self.fov_horiz,
            "vert": self.vert,
            "horiz": self.horiz,
            "normal": self.normal,
        }
        zone_data.update(data)
        return zone_data


class CalcRectangle(Plane):
    def __init__(
        self,
        x1=0,
        x2=6,
        y1=0,
        y2=4,
    ):

        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2


class Volume(Zone):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # forwards all flag kwargs

    def plot(self):
        """"""

    def export(self, fname):
        """"""


class CalcPlaneBase:
    def __init__(self, coords, normal, vert, horiz, fov_vert, fov_horiz):

        self.coords = coords
        points = [np.unique(val) for val in self.coords.T]
        self.num_points = np.array([len(val) for val in points if len(val) > 1])

        self.normal = normal
        self.basis = self._get_basis()

        self.vert = False if vert is None else vert
        self.horiz = False if horiz is None else horiz
        self.fov_vert = 180 if fov_vert is None else fov_vert
        self.fov_horiz = 360 if fov_horiz is None else abs(fov_horiz)

        # everything to do with calculation is here
        self.calculator = LightingCalculator(self)
        self.base_values = None
        self.reflected_values = np.zeros(self.num_points)
        self.values = np.zeros(self.num_points)

        self.lamp_values = {}
        self.lamp_values_base = {}

    def _get_coplanar(self, tol=1e-12):
        """
        Return the first three non‑collinear points.
        points : (N, 3) array_like
        tol    : float   – threshold for ‖v1 × v2‖ below which we call it collinear
        """
        combos = combinations(range(len(self.coords)), 3)
        for i, j, k in combos:
            v1 = self.coords[j] - self.coords[i]
            v2 = self.coords[k] - self.coords[i]
            if np.linalg.norm(np.cross(v1, v2)) > tol:
                return np.array((self.coords[i], self.coords[j], self.coords[k]))
        raise ValueError("All points collinear within tolerance")

    def _get_normal(self):
        """determine the plane's normal from the coordinates"""
        pts = self._get_coplanar()
        v1 = pts[1] - pts[0]
        v2 = pts[2] - pts[0]
        n = np.cross(v1, v2)
        return n / np.linalg.norm(n)

    def _get_basis(self, normal=None):
        # Generate arbitrary vector not parallel to n

        n = self._get_normal() if normal is None else normal
        tmp = np.array([1, 0, 0]) if abs(n[0]) < 0.9 else np.array([0, 1, 0])

        u = np.cross(n, tmp)
        u = u / np.linalg.norm(u)

        v = np.cross(n, u)
        return np.stack([u, v, n], axis=1)

    def _reset_params(self):
        self.normal = self._get_normal()
        self.vert = False
        self.horiz = False
        self.fov_vert = 180
        self.fov_horiz = 360

    def set_calc_type(self, calc_type, point=None, fov_vert=None, fov_horiz=None):
        """
        TODO:
        the eye calc type should be handled pretty differently in the calc manager eventually
        the directional
        """

        calc_types = [
            "planar_norm",
            "planar_max",
            "fluence",
            "eye",
            "directional",
            "target",
        ]
        calc_type = calc_type.lower()

        if calc_type not in calc_types:
            msg = f"{calc_type} is not a valid calculation type; valid calculation types are {calc_types}"
            raise KeyError(msg)

        if calc_type == "planar_norm":
            self._reset_params()
            self.horiz = True  # may not be necessary later
        elif calc_type == "planar_max":
            self._reset_params()
        elif calc_type == "fluence":
            self._reset_params()
            self.normal = None
        elif calc_type == "eye":
            self._reset_params()
            self.normal = None
            self.vert = True
            self.fov_vert = 80 if fov_vert is None else fov_vert
            self.fov_horiz = 180 if fov_horiz is None else fov_horiz
        else:
            if point is None:
                msg = "For the directional and target calculation types, a normal must be provided"
                raise ValueError(msg)
            elif len(point) != 3:
                msg = "Target point must be a cartesian point of len(3)"
                raise ValueError(msg)

            if calc_type == "directional":
                norm = point - self.coords.mean(axis=0)
                self.normal = norm / np.linalg.norm(norm)
                self.basis = self._get_basis(self.normal)
            elif calc_type == "target":
                norms = point - self.coords.astype("float32")
                norms = (norms.T / np.linalg.norm(norms, axis=1)).T
                self.normal = norms

    def _write_rows(self):
        """
        export solution to csv
        """

        num_x, num_y = self.num_points

        values = self.get_values()
        if self.base_values is None:
            vals = [[-1] * num_y] * num_x
        elif values.shape != (num_x, num_y):
            vals = [[-1] * num_y] * num_x
        else:
            vals = values
        zvals = self.coords.T[2].reshape(num_x, num_y).T[::-1]

        xpoints = self.coords.T[0].reshape(num_x, num_y).T[0].tolist()
        ypoints = self.coords.T[1].reshape(num_x, num_y)[0].tolist()

        if len(np.unique(xpoints)) == 1 and len(np.unique(ypoints)) == 1:
            xpoints = self.coords.T[0].reshape(num_x, num_y)[0].tolist()
            ypoints = self.coords.T[1].reshape(num_x, num_y).T[0].tolist()
            vals = vals.T
            zvals = zvals.T

        rows = [[""] + xpoints]
        rows += np.concatenate(([np.flip(ypoints)], vals)).T.tolist()
        rows += [""]
        rows += [[""] + list(line) for line in zvals]
        return rows

        def plot_plane(self, fig=None, ax=None, vmin=None, vmax=None, title=None):
            """Plot the image of the radiation pattern"""
            if fig is None:
                if ax is None:
                    fig, ax = plt.subplots()
                else:
                    fig = plt.gcf()
            else:
                if ax is None:
                    ax = fig.axes[0]

            title = "" if title is None else title
            values = self.get_values()
            xp, yp = [
                np.unique(val) for val in self.coords.T if len(np.unique(val)) > 1
            ]
            if values is not None:
                vmin = values.min() if vmin is None else vmin
                vmax = values.max() if vmax is None else vmax
                extent = [xp[0], xp[-1], yp[0], yp[-1]]
                values = values.T[::-1]
                img = ax.imshow(
                    values, extent=extent, vmin=vmin, vmax=vmax, cmap="plasma"
                )
                cbar = fig.colorbar(img, pad=0.03)
                ax.set_title(title)
                cbar.set_label(self.units, loc="center")
            return fig, ax



from importlib import resources
import inspect
import json
import pathlib
import warnings
import copy
import numpy as np
from photompy import Photometry, IESFile
from .spectrum import Spectrum
from .lamp_surface import LampSurface
from .lamp_plotter import LampPlotter
from .lamp_orientation import LampOrientation
from .trigonometry import to_polar
from ._data import get_tlvs

VALID_LAMPS = [
    "aerolamp",
    "beacon",
    "lumenizer_zone",
    "nukit_lantern",
    "nukit_torch",
    "sterilray",
    "ushio_b1",
    "ushio_b1.5",
    "uvpro222_b1",
    "uvpro222_b2",
    "visium",
]

KRCL_KEYS = [
    "krypton chloride",
    "krypton-chloride",
    "krypton_chloride",
    "krcl",
    "kr-cl",
    "kr cl",
]

LPHG_KEYS = [
    "low pressure mercury",
    "low-pressure mercury",
    "mercury",
    "lphg",
    "lp-hg",
    "lp hg",
]


class Lamp:
    """
    Represents a lamp with properties defined by a photometric data file.
    This class handles the loading of IES photometric data, orienting the lamp in 3D space,
    and provides methods for moving, rotating, and aiming the lamp.

    Arguments
    -------------------
    lamp_id: str
        A unique identifier for the lamp object.
    name: str, default=None
        Non-unique display name for the lamp. If None set by lamp_id
    filename: Path, str
        If None or not pathlike, `filedata` must not be None
    filedata: Path or bytes, default=None
        Set by `filename` if filename is pathlike.
    x, y, z: floats, default=[0,0,0]
        Sets initial position of lamp in cartesian space
    angle: float, default=0
        Sets lamps initial rotation on its own axis.
    aimx, aimy, aimz: floats, default=[0,0,z-1]
        Sets initial aim point of lamp in cartesian space.
    guv_type: str
        Optional label for type of GUV source. Presently available:
        ["Krypton chloride (222 nm)", "Low-pressure mercury (254 nm)", "Other"]
    wavelength: float
        Optional label for principle GUV wavelength. Set from guv_type if guv_type
        is not "Other".
    spectra_source: Path or bytes, default=None
        Optional. Data source for spectra. May be a filepath, a binary stream,
        or a dict where the first value contains values of wavelengths, and
        the second value contains values of relative intensity.
    width, length, depth: floats, default=[None, None]
        x-axis and y-axis source extent, plus fixture depth in the units
        provided. If not provided, will be read from the .ies file. Note that
        the ies file interface labels `depth` as `height` instead.
    units: str or int in [1, 2] or None
        `feet` or `meters`. 1 corresponds to feet, 2 to `meters`. If not
        provided, will be read from .ies file, and lengt and width parameters
        will be ignored.
    source_density: int or float, default=1
        parameter that determines the fineness of the source discretization.
        Grid size follows fibonacci sequence. For an approximately square
        source, SD=1 => 1x1 grid, SD=2 => 3x3 grid, SD=3 => 5x5 grid. This is
        to ensure that a center point is always present while ensuring evenness
        of grid size.
    intensity_map: arraylike
        A relative intensity map for non-uniform sources. Must be of the same
        size as the grid generated
    enabled: bool, defualt=True
        Determines if lamp participates in calculations. A lamp may be created
        and added to a room, but disabled.

    """

    def __init__(
        self,
        lamp_id=None,
        name=None,
        filedata=None,
        filename=None,
        x=None,
        y=None,
        z=None,
        angle=None,
        aimx=None,
        aimy=None,
        aimz=None,
        intensity_units=None,
        guv_type=None,
        wavelength=None,
        spectra_source=None,
        width=None,
        length=None,
        depth=None,
        units=None,
        source_density=None,
        intensity_map=None,
        enabled=None,
        scaling_factor=None,
    ):

        self.lamp_id = lamp_id
        self.name = str(lamp_id) if name is None else str(name)
        self.enabled = True if enabled is None else enabled

        # Position / orientation
        x = 0.0 if x is None else x
        y = 0.0 if y is None else y
        z = 0.0 if z is None else z
        self.pose = LampOrientation(
            x=x,
            y=y,
            z=z,
            angle=0.0 if angle is None else angle,
            aimx=x if aimx is None else aimx,
            aimy=y if aimy is None else aimy,
            aimz=z - 1.0 if aimz is None else aimz,
        )

        # Surface data
        self.surface = LampSurface(
            width=width,
            length=length,
            depth=depth,
            units=units,
            source_density=source_density,
            intensity_map=intensity_map,
            pose=self.pose,
        )

        # Photometric data
        if filename is not None:
            warnings.warn(
                "`filename` is deprecated and will be removed in v1.4; "
                "pass the bytes or Path in `filedata` instead.",
                DeprecationWarning,  # maybe should be FutureWarning
                stacklevel=2,
            )
            # honour old behaviour if the new arg wasn’t supplied
            if filedata is None:
                filedata = filename
        self.filedata = filedata  # temp - property eventually to be removed
        self.ies = None
        self._base_ies = None
        self._scaling_factor = scaling_factor or 1.0
        self._scale_mode = "factor"
        self.load_ies(filedata)
        self.filename = None  # VERY temp - just for illluminate compatibility

        # Spectral data
        self.spectra_source = spectra_source
        self.spectra = self._load_spectra(spectra_source)

        # source type & wavelength - just labels if Spectra is provided
        self.guv_type = guv_type
        if guv_type is not None:
            if any([key in guv_type.lower() for key in KRCL_KEYS]):
                self.wavelength = 222
            elif any([key in guv_type.lower() for key in LPHG_KEYS]):
                self.wavelength = 254
            else:  # set from wavelength if guv_type not recognized
                self.wavelength = wavelength
        else:
            self.wavelength = wavelength
        if self.wavelength is not None:
            if not isinstance(self.wavelength, (int, float)):
                msg = f"Wavelength must be int or float, not {type(self.wavelength)}"
                raise TypeError(msg)

        # mW/sr or uW/cm2 typically; not directly specified in .ies file and they can vary for GUV fixtures
        self.intensity_units = self._set_intensity_units(intensity_units)

        # plotting
        self.plotter = LampPlotter(self)

        # state
        self.calc_state = None
        self.update_state = None

    # ------------------------ Basics ------------------------------

    def to_dict(self):
        """
        save just the minimum number of parameters required to re-instantiate the lamp
        Returns dict. If filename is not None, saves dict as json.
        """

        data = {}
        data["lamp_id"] = self.lamp_id
        data["name"] = self.name
        data["x"] = float(self.pose.x)
        data["y"] = float(self.pose.y)
        data["z"] = float(self.pose.z)
        data["angle"] = float(self.pose.angle)
        data["aimx"] = float(self.pose.aimx)
        data["aimy"] = float(self.pose.aimy)
        data["aimz"] = float(self.pose.aimz)
        data["intensity_units"] = self.intensity_units
        data["guv_type"] = self.guv_type
        data["wavelength"] = self.wavelength
        data["width"] = self.surface.width
        data["length"] = self.surface.length
        data["depth"] = self.surface.depth
        data["units"] = self.surface.units
        data["source_density"] = self.surface.source_density
        data["scaling_factor"] = float(self.scaling_factor)

        if self.surface.intensity_map_orig is None:
            data["intensity_map"] = None
        else:
            data["intensity_map"] = self.surface.intensity_map_orig.tolist()

        data["enabled"] = True

        data["filename"] = str(self.filename)
        filedata = self.save_ies(original=True)
        data["filedata"] = filedata.decode() if filedata is not None else None

        if self.spectra is not None:
            spectra_dict = self.spectra.to_dict(as_string=True)
            keys = list(spectra_dict.keys())[0:2]  # keep the first two keys only
            data["spectra"] = {key: spectra_dict[key] for key in keys}
        else:
            data["spectra"] = None
        return data

    @classmethod
    def from_dict(cls, data):
        """initialize class from dict"""
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        if data["spectra"] is not None:
            data["spectra_source"] = {}
            for k, v in data["spectra"].items():
                if isinstance(v, str):
                    lst = list(map(float, v.split(", ")))
                elif isinstance(v, list):
                    lst = v
                data["spectra_source"][k] = np.array(lst)
        return cls(**{k: v for k, v in data.items() if k in keys})

    @property
    def keywords(self):
        return VALID_LAMPS

    @classmethod
    def from_keyword(cls, key, **kwargs):
        """define a Lamp object from a predefined keyword"""
        if not isinstance(key, str):
            raise TypeError(f"Keyword must be str, not {type(key)}")
        if key.lower() in VALID_LAMPS:
            path = "guv_calcs.data.lamp_data"
            fn = resources.files(path).joinpath(key.lower() + ".ies")
            sn = resources.files(path).joinpath(key.lower() + ".csv")
            kwargs.setdefault("filedata", fn)
            kwargs.setdefault("spectra_source", sn)
        else:
            raise KeyError(
                f"{key} is not a valid lamp key. Valid keys are {VALID_LAMPS}"
            )
        if kwargs.get("lamp_id", None) is None:
            kwargs.setdefault("lamp_id", key)

        return cls(**kwargs)

    @classmethod
    def from_index(cls, key_index=0, **kwargs):
        """define a Lamp object from an index value"""

        if not isinstance(key_index, int):
            raise TypeError(f"Keyword index must be int, not {type(key_index)}")

        if key_index < len(VALID_LAMPS):
            key = VALID_LAMPS[key_index]
            path = "guv_calcs.data.lamp_data"
            fn = resources.files(path).joinpath(key.lower() + ".ies")
            sn = resources.files(path).joinpath(key.lower() + ".csv")
            kwargs.setdefault("filedata", fn)
            kwargs.setdefault("spectra_source", sn)
        else:
            raise IndexError(
                f"Only {len(VALID_LAMPS)} lamps are available. Available lamps: {VALID_LAMPS}"
            )
        if kwargs.get("lamp_id", None) is None:
            kwargs.setdefault("lamp_id", key)

        return cls(**kwargs)

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this lamp must be recalculated
        """
        # this needs summed to make comparison not fail, might need to investigate later
        intensity_map_orig = (
            self.surface.intensity_map_orig.sum()
            if self.surface.intensity_map_orig is not None
            else None
        )
        return [
            self.filedata,
            self.x,
            self.y,
            self.z,
            self.angle,
            self.aimx,
            self.aimy,
            self.aimz,
            self.surface.length,  # only for nearfield
            self.surface.width,  # ""
            self.surface.depth,
            self.surface.units,  # ""
            self.surface.source_density,  # ""
            intensity_map_orig,
            self.scaling_factor,
        ]

    def get_update_state(self):
        return [self.intensity_units]

    # ----------------------- IO ------------------------------------

    def load_ies(self, filedata, override=True):
        """load an ies file"""

        self.filedata = filedata  # tmp
        if filedata is None:
            self._base_ies = None
        elif isinstance(filedata, IESFile):
            self._base_ies = filedata
        elif isinstance(filedata, Photometry):
            self._base_ies = IESFile.from_photometry(filedata)
        else:  # all other datasource cases covered here
            self._base_ies = IESFile.read(filedata)

        # in case the base object is mutated
        self.ies = copy.deepcopy(self._base_ies)

        if self.ies is not None:
            self.ies.scale(self.scaling_factor)

        # update length/width/units
        if override:
            self.surface.set_ies(self.ies)

        return self.ies

    def load_spectra(self, spectra_source):
        """external method for loading spectra after lamp object has been instantiated"""
        self.spectra = self._load_spectra(spectra_source)

    def load_intensity_map(self, intensity_map):
        """external method for loading relative intensity map after lamp object has been instantiated"""
        self.surface.load_intensity_map(intensity_map)

    def save_ies(self, fname=None, original=False):
        """
        Save the current lamp paramters as an .ies file; alternatively, save the original ies file.
        """
        if self.ies is not None:
            if original:
                iesbytes = self._base_ies.write(which="orig")
            else:
                iesbytes = self.ies.write(which="orig")

            # write to file if provided, otherwise
            if fname is not None:
                with open(fname, "wb") as file:
                    file.write(iesbytes)
            else:
                return iesbytes
        else:
            return None

    def save(self, filename):
        """save lamp information as json"""
        data = self.to_dict()
        with open(filename, "w") as json_file:
            json.dump(data, json_file, indent=4)
        return data

    def copy(self, lamp_id):
        """copy the lamp object with a new ID"""
        lamp = copy.deepcopy(self)
        lamp.lamp_id = lamp_id
        return lamp

    # ------------------- Position / Orientation ---------------------

    # temp properties...
    @property
    def x(self):
        return self.pose.x

    @property
    def y(self):
        return self.pose.y

    @property
    def z(self):
        return self.pose.z

    @property
    def position(self):
        return self.pose.position

    @property
    def aimx(self):
        return self.pose.aimx

    @property
    def aimy(self):
        return self.pose.aimy

    @property
    def aimz(self):
        return self.pose.aimz

    @property
    def aim_point(self):
        return self.pose.aim_point

    @property
    def angle(self):
        return self.pose.angle

    @property
    def heading(self):
        return self.pose.heading

    @property
    def bank(self):
        return self.pose.bank

    def move(self, x=None, y=None, z=None):
        """Designate lamp position in cartesian space"""
        self.pose = self.pose.move(x=x, y=y, z=z)
        self.surface.set_pose(self.pose)
        return self

    def rotate(self, angle):
        """designate lamp orientation with respect to its z axis"""
        self.pose = self.pose.rotate(angle)
        self.surface.set_pose(self.pose)
        return self

    def aim(self, x=None, y=None, z=None):
        """aim lamp at a point in cartesian space"""
        self.pose = self.pose.aim(x=x, y=y, z=z)
        self.surface.set_pose(self.pose)
        return self

    def transform_to_world(self, coords, scale=1, which="cartesian"):
        """
        transform coordinates from the lamp frame of reference to the world
        Scale parameter should generally only be used for photometric_coords
        """
        return self.pose.transform_to_world(coords, scale=scale, which=which)

    def transform_to_lamp(self, coords, which="cartesian"):
        """
        transform coordinates to align with the lamp's coordinates
        """
        return self.pose.transform_to_lamp(coords, which=which)

    def set_orientation(self, orientation, dimensions=None, distance=None):
        """
        set orientation/heading.
        alternative to setting aim point with `aim`
        distinct from rotation; applies to a tilted lamp. to rotate a lamp along its axis,
        use the `rotate` method
        """
        self.pose = self.pose.recalculate_aim_point(
            heading=orientation, dimensions=dimensions, distance=distance
        )
        self.surface.set_pose(self.pose)
        return self

    def set_tilt(self, tilt, dimensions=None, distance=None):
        """
        set tilt/bank
        alternative to setting aim point with `aim`
        """
        self.pose = self.pose.recalculate_aim_point(
            bank=tilt, dimensions=dimensions, distance=distance
        )
        self.surface.set_pose(self.pose)
        return self

    # ---------------------- Photometry --------------------------------

    @property
    def photometry(self) -> Photometry | None:
        """Active Photometry block (or None if lamp has no photometry)."""
        if self.ies is None:
            return None
        return self.ies.photometry

    @property
    def thetas(self):
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        return self.ies.photometry.expanded().thetas

    @property
    def phis(self):
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        return self.ies.photometry.expanded().phis

    @property
    def values(self):
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        return self.ies.photometry.expanded().values

    @property
    def coords(self):
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        return self.ies.photometry.coords

    @property
    def photometric_coords(self):
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        return self.ies.photometry.photometric_coords

    def max(self):
        """maximum irradiance value"""
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        if self.intensity_units == "mW/sr":
            return self.ies.photometry.max() / 10
        else:
            return self.ies.photometry.max()

    def center(self):
        """center irradiance value"""
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        if self.intensity_units == "mW/sr":
            return self.ies.photometry.center() / 10
        else:
            return self.ies.photometry.center()

    def total(self):
        """just an alias for get_total_power for now"""
        return self.get_total_power()

    def get_total_power(self):
        """return the lamp's total optical power"""
        if self.ies is None:
            raise AttributeError("Lamp has no photometry")
        if self.intensity_units == "mW/sr":
            return self.ies.photometry.total_optical_power()
        else:
            return self.ies.photometry.total_optical_power() * 10

    def get_tlvs(self, standard=0):
        """
        get the threshold limit values for this lamp. Returns tuple
        (skin_limit, eye_limit) Will use the lamp spectrum if provided;
        if not provided will use wavelength; if neither is defined, returns
        (None, None). Standard may be a string in:
            [`ANSI IES RP 27.1-22`, `IEC 62471-6:2022`]
        Or an integer corresponding to the index of the desired standard.
        """
        if self.spectra is not None:
            skin_tlv, eye_tlv = get_tlvs(self.spectra, standard)
        elif self.wavelength is not None:
            skin_tlv, eye_tlv = get_tlvs(self.wavelength, standard)
        else:
            skin_tlv, eye_tlv = None, None
        return skin_tlv, eye_tlv

    def get_limits(self, standard=0):
        """compatibility alias for `get_tlvs()`"""
        return self.get_tlvs(standard=standard)

    def get_cartesian(self, scale=1, sigfigs=9):
        """Return lamp's true position coordinates in cartesian space"""
        return self.transform(self.coords, scale=scale).round(sigfigs)

    def get_polar(self, sigfigs=9):
        """Return lamp's true position coordinates in polar space"""
        cartesian = self.transform(self.coords) - self.position
        return np.array(to_polar(*cartesian.T)).round(sigfigs)

    # ---- scaling / dimming features -----

    @property
    def scaling_factor(self) -> float:
        """Current multiplier relative to the loaded photometry."""
        return self._scaling_factor

    @scaling_factor.setter  # block direct writes
    def scaling_factor(self, _):
        raise AttributeError("scaling_factor is read-only")

    def scale(self, scale_val):
        """scale the photometry by the given value"""
        if self.ies is None:
            msg = "No .ies file provided; scaling not applied"
            warnings.warn(msg, stacklevel=3)
        else:
            self.photometry.scale(scale_val / self.scaling_factor)
            self._update_scaling_factor()
        return self

    def scale_to_max(self, max_val):
        """scale the photometry to a maximum value [in uW/cm2]"""
        if self.ies is None:
            msg = "No .ies file provided; scaling not applied"
            warnings.warn(msg, stacklevel=3)
        else:
            if self.intensity_units == "mW/sr":
                self.ies.photometry.scale_to_max(max_val * 10)
            else:
                self.ies.photometry.scale_to_max(max_val)
            self._update_scaling_factor()
        return self

    def scale_to_total(self, total_power):
        """scale the photometry to a total optical power [in mW]"""
        if self.ies is None:
            msg = "No .ies file provided; scaling not applied"
            warnings.warn(msg, stacklevel=3)
        else:
            self.ies.photometry.scale_to_total(total_power)
            self._update_scaling_factor()
        return self

    def scale_to_center(self, center_val):
        """scale the photometry to a center irradiance value [in uW/cm2]"""
        if self.ies is None:
            msg = "No .ies file provided; scaling not applied"
            warnings.warn(msg, stacklevel=3)
        else:
            if self.intensity_units == "mW/sr":
                self.photometry.scale_to_center(center_val * 10)
            else:
                self.photometry.scale_to_center(center_val)
            self._update_scaling_factor()
            # self._scale_mode = "center"
        return self

    def _update_scaling_factor(self):
        """update scaling factor based on the last scaling operation"""
        self._scaling_factor = self.ies.center() / self._base_ies.center()

    # ---------------------- Surface ---------------------------

    @property
    def units(self):
        return self.surface.units

    @property
    def length(self):
        return self.surface.length

    @property
    def width(self):
        return self.surface.width

    @property
    def depth(self):
        return self.surface.depth

    def set_source_density(self, source_density):
        """change source discretization"""
        self.surface.set_source_density(source_density)

    def set_units(self, units):
        """set units"""
        if self.ies is not None:
            self.ies.update(units=1 if units == "feet" else 2)
        self.surface.set_units(units)
        return self

    def set_width(self, width):
        """change x-axis extent of lamp emissive surface"""
        self.ies.update(width=width)
        self.surface.set_width(width)
        return self

    def set_length(self, length):
        """change y-axis extent of lamp emissive surface"""
        self.ies.update(length=length)
        self.surface.set_length(length)
        return self

    def set_depth(self, depth):
        """
        TODO: this should actually be decoupled from the ies file, that
        property is rightly height not depth.
        change the z-axis offset of where the lamp's emissive surface is
        """
        self.ies.update(depth=depth)
        self.surface.set_depth(depth)

    # ------------------------ Plotting ------------------------------

    def plot_ies(self, **kwargs):
        """see LampPlotter.plot_ies"""
        return self.plotter.plot_ies(**kwargs)

    def plot_web(self, **kwargs):
        """see LampPlotter.plot_web"""
        return self.plotter.plot_web(**kwargs)

    def plot_3d(self, **kwargs):
        """see LampPlotter.plot_3d"""
        return self.plotter.plot_3d(**kwargs)

    def plot_spectra(self, **kwargs):
        """see LampPlotter.plot_spectra and Spectrum.plot"""
        return self.plotter.plot_spectra(**kwargs)

    def plot_surface(self, **kwargs):
        """see LampSurface.plot_surface"""
        return self.surface.plot_surface(**kwargs)

    # --------------------------- Internals -----------------------------

    def _set_intensity_units(self, arg):
        """
        TODO: this should probably just be an enum?
        determine the units of the radiant intensity
        """
        if arg is not None:
            msg = f"Intensity unit {arg} not recognized. Using default value `mW/sr`"
            if isinstance(arg, int):
                if arg == 0:
                    intensity_units = "mW/sr"
                elif arg == 1:
                    intensity_units = "uW/cm²"
                else:
                    warnings.warn(msg, stacklevel=3)
                    intensity_units = "mW/sr"
            elif isinstance(arg, str):
                if arg.lower() in ["mw/sr", "uw/cm2", "uw/cm²"]:
                    intensity_units = arg
                else:
                    warnings.warn(msg, stacklevel=3)
                    intensity_units = "mW/sr"
            else:
                msg = f"Datatype {type(arg)} for intensity units not recognized. Using default value `mW/Sr`"
                intensity_units = "mW/sr"
        else:
            intensity_units = "mW/sr"
        return intensity_units

    def _load_spectra(self, spectra_source):
        """initialize a Spectrum object from the source"""
        if isinstance(spectra_source, dict):
            spectra = Spectrum.from_dict(spectra_source)
        elif isinstance(spectra_source, (str, pathlib.Path, bytes)):
            spectra = Spectrum.from_file(spectra_source)
        elif isinstance(spectra_source, tuple):
            spectra = Spectrum(spectra_source[0], spectra_source[1])
        elif spectra_source is None:
            spectra = None
        else:
            spectra = None
            msg = f"Datatype {type(spectra_source)} not recognized spectral data source"
            warnings.warn(msg, stacklevel=3)
        return spectra



import inspect
import warnings
import json
import copy
import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objs as go
from .calc_manager import LightingCalculator
from .io import rows_to_bytes


class CalcZone(object):
    """
    TODO: dummy this out?

    Base class representing a calculation zone.

    This class provides a template for setting up zones within which various
    calculations related to lighting conditions are performed. Subclasses should
    provide specific implementations of the coordinate setting method.

    NOTE: I changed this from an abstract base class to an object superclass
    to make it more convenient to work with the website, but this class doesn't really
    work on its own

    Parameters:
    --------
    zone_id: str
        identification tag for internal tracking
    name: str, default=None
        externally visible name for zone
    offset: bool, default=True
    dose: bool, default=False
        whether to calculate a dose over N hours or just fluence
    hours: float, default = 8.0
        number of hours to calculate dose over. Only relevant if dose is True.
    enabled: bool, default = True
        whether or not the calc zone is enabled for calculations
    """

    def __init__(
        self,
        zone_id=None,
        name=None,
        offset=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
        colormap=None,
    ):
        self.zone_id = zone_id
        self.name = str(zone_id) if name is None else str(name)
        self.offset = True if offset is None else offset
        self.dose = False if dose is None else dose
        if self.dose:
            self.units = "mJ/cm²"
        else:
            self.units = "uW/cm²"
        self.hours = 8.0 if hours is None else abs(hours)  # only used if dose is true
        self.enabled = True if enabled is None else enabled
        self.show_values = True if show_values is None else show_values
        self.colormap = colormap

        self.calculator = LightingCalculator(self)

        # these will all be calculated after spacing is set, which is set in the subclass
        self.calctype = "Zone"
        self.x1 = None
        self.x2 = None
        self.y1 = None
        self.y2 = None
        self.z1 = None
        self.z2 = None
        self.height = None
        self.num_x = None
        self.num_y = None
        self.num_z = None
        self.x_spacing = None
        self.y_spacing = None
        self.z_spacing = None
        self.num_points = None
        self.xp = None
        self.yp = None
        self.zp = None
        self.coords = None
        self.ref_surface = None
        self.direction = None
        self.horiz = None
        self.vert = None
        self.fov_vert = None
        self.fov_horiz = None
        self.basis = None
        self.values = None
        self.reflected_values = None
        self.lamp_values = {}
        self.lamp_values_base = {}
        self.calc_state = None

    def to_dict(self):

        data = {}
        data["zone_id"] = self.zone_id
        data["name"] = self.name
        data["offset"] = self.offset
        data["fov_vert"] = self.fov_vert
        data["fov_horiz"] = self.fov_horiz
        data["vert"] = self.vert
        data["horiz"] = self.horiz
        data["dose"] = self.dose
        data["hours"] = self.hours
        data["enabled"] = self.enabled
        data["show_values"] = self.show_values
        data["x1"] = self.x1
        data["x2"] = self.x2
        data["x_spacing"] = self.x_spacing
        # data["num_x"] = self.num_x
        data["y1"] = self.y1
        data["y2"] = self.y2
        data["y_spacing"] = self.y_spacing
        # data["num_y"] = self.num_y
        if isinstance(self, CalcPlane):
            data["height"] = self.height
            data["calctype"] = "Plane"
        elif isinstance(self, CalcVol):
            data["z1"] = self.z1
            data["z2"] = self.z2
            data["z_spacing"] = self.z_spacing
            # data["num_z"] = self.num_z
            data["calctype"] = "Volume"
        return data

    def save(self, filename):
        """save zone data to json file"""
        data = self.to_dict()
        with open(filename, "w") as json_file:
            json_file.write(json.dumps(data))

    @classmethod
    def from_dict(cls, data):
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        return cls(**{k: v for k, v in data.items() if k in keys})

    def set_dimensions(self, dimensions):
        raise NotImplementedError

    def set_spacing(self, spacing):
        raise NotImplementedError

    def _write_rows(self):
        raise NotImplementedError

    def set_offset(self, offset):
        if type(offset) is not bool:
            raise TypeError("must be either True or False")
        self.offset = offset
        self._update()

    def set_value_type(self, dose):
        """
        if true get_values() will return in dose (mJ/cm2) over hours
        otherwise get_values() will return in irradiance or fluence (uW/cm2)
        """
        if type(dose) is not bool:
            raise TypeError("Dose must be either True or False")

        # # convert values if they need converting
        # if self.values is not None:
        # if dose and not self.dose:
        # self.values = self.values * 3.6 * self.hours
        # elif self.dose and not dose:
        # self.values = self.values / (3.6 * self.hours)

        self.dose = dose
        if self.dose:
            self.units = "mJ/cm²"
        else:
            self.units = "uW/cm²"

    def set_dose_time(self, hours):
        """
        Set the time over which the dose will be calculate in hours
        """
        if type(hours) not in [float, int]:
            raise TypeError("Hours must be numeric")
        self.hours = hours

        # if self.values is not None and self.dose:
        # self.values = (self.values * hours) / self.hours

    def _set_spacing(self, pt1, pt2, num, spacing):
        """set the spacing value conservatively from a num_points value"""
        rnge = abs(pt2 - pt1)
        if int(rnge / spacing) == int(num):
            val = spacing  # no changes needed
        else:
            testval = rnge / round(num)
            i = 1
            while i < 6:

                val = round(testval, i)
                if val != 0 and int(rnge / round(testval, i) + 1) == num:
                    break
                i += 1
                val = testval  # if no rounded value works use the original value
        return val

    def calculate_values(self, lamps, ref_manager=None, hard=False):
        """
        Calculate all the values for all the lamps
        """

        new_calc_state = self.get_calc_state()

        # updates self.lamp_values_base and self.lamp_values
        self.base_values = self.calculator.compute(lamps=lamps, hard=hard)

        if ref_manager is not None:
            # calculate reflectance -- warning, may be expensive!
            ref_manager.calculate_reflectance(self, hard=hard)
            # add in reflected values, if applicable
            self.reflected_values = ref_manager.get_total_reflectance(self)
        else:
            self.reflected_values = np.zeros(self.num_points).astype("float32")

        # sum
        self.values = self.base_values + self.reflected_values
        self.calc_state = new_calc_state

        return self.get_values()

    def get_values(self):
        """
        return
        """
        if self.values is None:
            values = None
        else:
            if self.dose:
                values = self.values * 3.6 * self.hours
            else:
                values = self.values
        return values

    def export(self, fname=None):
        """
        export the calculation zone's results to a .csv file
        if the spacing has been updated but the values not recalculated,
        exported values will be blank.
        """
        try:
            rows = self._write_rows()  # implemented in subclass
            csv_bytes = rows_to_bytes(rows)

            if fname is not None:
                with open(fname, "wb") as csvfile:
                    csvfile.write(csv_bytes)
            else:
                return csv_bytes
        except NotImplementedError:
            pass

    def copy(self, zone_id):
        """
        return a copy of this CalcZone with the same attributes and a new zone_id
        """
        zone = copy.deepcopy(self)
        zone.zone_id = zone_id
        # clear calculated values
        zone.values = None
        zone.reflected_values = None
        zone.lamp_values = {}
        zone.lamp_values_base = {}
        return zone


class CalcVol(CalcZone):
    """
    Represents a volumetric calculation zone.
    A subclass of CalcZone designed for three-dimensional volumetric calculations.
    """

    def __init__(
        self,
        zone_id=None,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        z1=None,
        z2=None,
        num_x=None,
        num_y=None,
        num_z=None,
        x_spacing=None,
        y_spacing=None,
        z_spacing=None,
        offset=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):

        super().__init__(
            zone_id=zone_id or "CalcVol",
            name=name,
            offset=offset,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Volume"
        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.z1 = 0 if z1 is None else z1
        self.z2 = 2.7 if z2 is None else z2
        self.dimensions = ((self.x1, self.y1, self.z1), (self.x2, self.y2, self.z2))

        # spacing and num points
        xr = abs(self.x2 - self.x1)
        yr = abs(self.y2 - self.y1)
        zr = abs(self.z2 - self.z1)
        nx = min(max(int(xr * 10), 1), 20)
        ny = min(max(int(yr * 10), 1), 20)
        nz = min(max(int(zr * 10), 1), 20)
        default_xs = round(xr / nx, -int(np.floor(np.log10(xr / nx))))
        default_ys = round(yr / ny, -int(np.floor(np.log10(yr / ny))))
        default_zs = round(zr / nz, -int(np.floor(np.log10(zr / nz))))
        default_num_x = int(round(xr / default_xs))
        default_num_y = int(round(yr / default_ys))
        default_num_z = int(round(zr / default_zs))
        if x_spacing is None:  # set from num_points if spacing is not provided
            self.num_x = default_num_x if num_x is None else int(num_x)
            self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, default_xs)
        else:
            self.x_spacing = x_spacing
            self.num_x = int(round(xr / self.x_spacing))
        if y_spacing is None:
            self.num_y = default_num_y if num_y is None else int(num_y)
            self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, default_ys)
        else:
            self.y_spacing = y_spacing
            self.num_y = int(round(yr / self.y_spacing))
        if z_spacing is None:
            self.num_z = default_num_z if num_z is None else int(num_z)
            self.z_spacing = self._set_spacing(self.z1, self.z2, self.num_z, default_zs)
        else:
            self.z_spacing = z_spacing
            self.num_z = int(round(zr / self.z_spacing))

        self._update()
        self.values = np.zeros(self.num_points).astype("float32")
        self.reflected_values = np.zeros(self.num_points).astype("float32")

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this calc zone must be recalculated
        """
        return [
            self.offset,
            self.x1,
            self.x2,
            self.x_spacing,
            self.y1,
            self.y2,
            self.y_spacing,
            self.z1,
            self.z2,
            self.z_spacing,
        ]

    def get_update_state(self):
        """
        return a set of parameters that, if changed, indicate that the
        calc zone need not be be recalculated, but may need updating
        Currently there are no relevant update parameters for a calc volume
        """
        return []

    def set_dimensions(
        self,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        z1=None,
        z2=None,
        preserve_spacing=True,
    ):
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self.z1 = self.z1 if z1 is None else z1
        self.z2 = self.z2 if z2 is None else z2
        self.dimensions = ((self.x1, self.y1, self.z1), (self.x2, self.y2, self.z2))

        # update number of points, keeping spacing
        xr = abs(self.x2 - self.x1)
        yr = abs(self.y2 - self.y1)
        zr = abs(self.z2 - self.z1)
        if preserve_spacing:
            self.num_x = max(int(round(xr / self.x_spacing)), 1)
            self.num_y = max(int(round(yr / self.y_spacing)), 1)
            self.num_z = max(int(round(zr / self.z_spacing)), 1)
        else:
            deg_x = -int(np.floor(np.log10(xr / self.num_x)))
            deg_y = -int(np.floor(np.log10(yr / self.num_y)))
            deg_z = -int(np.floor(np.log10(zr / self.num_z)))
            default_xs = round(xr / self.num_x, deg_x)
            default_ys = round(yr / self.num_y, deg_y)
            default_zs = round(zr / self.num_z, deg_z)
            self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, default_xs)
            self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, default_ys)
            self.z_spacing = self._set_spacing(self.z1, self.z2, self.num_z, default_zs)

        self._update()

    def set_spacing(self, x_spacing=None, y_spacing=None, z_spacing=None):
        """
        set the spacing desired in the dimension
        """
        self.x_spacing = self.x_spacing if x_spacing is None else abs(x_spacing)
        self.y_spacing = self.y_spacing if y_spacing is None else abs(y_spacing)
        self.z_spacing = self.z_spacing if z_spacing is None else abs(z_spacing)
        numx = int(round(abs(self.x2 - self.x1) / self.x_spacing))
        numy = int(round(abs(self.y2 - self.y1) / self.y_spacing))
        numz = int(round(abs(self.z2 - self.z1) / self.z_spacing))
        self.num_x = numx
        self.num_y = numy
        self.num_z = numz
        self._update()

    def set_num_points(self, num_x=None, num_y=None, num_z=None):
        """
        set the number of points desired in a dimension, instead of setting the spacing
        """
        self.num_x = self.num_x if num_x is None else abs(int(num_x))
        self.num_y = self.num_y if num_y is None else abs(int(num_y))
        self.num_z = self.num_z if num_z is None else abs(int(num_z))
        if self.num_x == 0:
            warnings.warn("Number of x points must be at least 1")
            self.num_x += 1
        if self.num_y == 0:
            warnings.warn("Number of y points must be at least 1")
            self.num_y += 1
        if self.num_z == 0:
            warnings.warn("Number of z points must be at least 1")
            self.num_z += 1

        # update spacing if required
        self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, self.x_spacing)
        self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, self.y_spacing)
        self.z_spacing = self._set_spacing(self.z1, self.z2, self.num_z, self.z_spacing)

        self._update()
        return self

    def _update(self):
        """
        Update the number of points based on the spacing, and then the points
        """

        if self.x1 == self.x2:
            self.num_x = 1
        if self.y1 == self.y2:
            self.num_y = 1
        if self.z1 == self.z2:
            self.num_z = 1

        x_offset = min(self.x1, self.x2)
        y_offset = min(self.y1, self.y2)
        z_offset = min(self.z1, self.z2)
        xp = np.array([i * self.x_spacing + x_offset for i in range(self.num_x)])
        yp = np.array([i * self.y_spacing + y_offset for i in range(self.num_y)])
        zp = np.array([i * self.z_spacing + z_offset for i in range(self.num_z)])

        if self.offset:
            xp += (abs(self.x2 - self.x1) - abs(xp[-1] - xp[0])) / 2
            yp += (abs(self.y2 - self.y1) - abs(yp[-1] - yp[0])) / 2
            zp += (abs(self.z2 - self.z1) - abs(zp[-1] - zp[0])) / 2

        self.xp, self.yp, self.zp = xp, yp, zp
        self.points = [self.xp, self.yp, self.zp]

        X, Y, Z = [
            grid.reshape(-1) for grid in np.meshgrid(*self.points, indexing="ij")
        ]
        self.coords = np.array((X, Y, Z)).T
        self.coords = np.unique(self.coords, axis=0)

        self.num_points = np.array([len(self.xp), len(self.yp), len(self.zp)])

    def _write_rows(self):
        """
        export solution to csv file
        designed to be in the same format as the Acuity Visual export
        """

        header = """Data format notes:

         Data consists of numZ horizontal grids of fluence rate values; each grid contains numX by numY points.

         numX; numY; numZ are given on the first line of data.
         The next line contains numX values; indicating the X-coordinate of each grid column.
         The next line contains numY values; indicating the Y-coordinate of each grid row.
         The next line contains numZ values; indicating the Z-coordinate of each horizontal grid.
         A blank line separates the position data from the first horizontal grid of fluence rate values.
         A blank line separates each subsequent horizontal grid of fluence rate values.

         fluence rate values are given in µW/cm²
         
         """
        lines = header.split("\n")
        rows = [[line] for line in lines]
        rows += [self.num_points]
        rows += self.points
        values = self.get_values()
        for i in range(self.num_z):
            rows += [""]
            if values is None:
                rows += [[""] * self.num_x] * self.num_y
            elif values.shape != (self.num_x, self.num_y, self.num_z):
                rows += [[""] * self.num_x] * self.num_y
            else:
                rows += values.T[i].tolist()
        return rows

    def plot_volume(
        self,
        title=None,
    ):
        """
        Plot the fluence values as an isosurface using Plotly.
        """

        if self.values is None:
            raise ValueError("No values calculated for this volume.")

        X, Y, Z = np.meshgrid(*self.points, indexing="ij")
        x, y, z = X.flatten(), Y.flatten(), Z.flatten()
        values = self.values.flatten()
        isomin = self.values.mean() / 2
        fig = go.Figure()
        fig.add_trace(
            go.Isosurface(
                x=x,
                y=y,
                z=z,
                value=values,
                isomin=isomin,
                surface_count=3,
                opacity=0.25,
                showscale=False,
                colorbar=None,
                colorscale=self.colormap,
                caps=dict(x_show=False, y_show=False, z_show=False),
                name=self.name + " Values",
            )
        )
        fig.update_layout(
            title=dict(
                text=self.name if title is None else title,
                x=0.5,  # center horizontally
                y=0.85,  # lower this value to move the title down (default is 0.95)
                xanchor="center",
                yanchor="top",
                font=dict(size=18),
            ),
            scene=dict(
                xaxis_title="x", yaxis_title="y", zaxis_title="z", aspectmode="data"
            ),
            height=450,
        )
        fig.update_scenes(camera_projection_type="orthographic")
        return fig


class CalcPlane(CalcZone):
    """
    Represents a planar calculation zone.
    A subclass of CalcZone designed for two-dimensional planar calculations at a specific height.
    """

    def __init__(
        self,
        zone_id=None,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        height=None,
        ref_surface="xy",
        direction=None,
        num_x=None,
        num_y=None,
        x_spacing=None,
        y_spacing=None,
        offset=None,
        fov_vert=None,
        fov_horiz=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):

        super().__init__(
            zone_id=zone_id or "CalcPlane",
            name=name,
            offset=offset,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Plane"
        self.height = 1.9 if height is None else height

        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.dimensions = ((self.x1, self.y1), (self.x2, self.y2))

        # TODO: this should really be much cleaner
        # perhaps just have users define cartesian points + a normal?
        # that may just need to be a slightly different class
        if not isinstance(ref_surface, str):
            raise TypeError("ref_surface must be a string in [`xy`,`xz`,`yz`]")
        if ref_surface.lower() not in ["xy", "xz", "yz"]:
            raise ValueError("ref_surface must be a string in [`xy`,`xz`,`yz`]")

        self.ref_surface = "xy" if ref_surface is None else ref_surface.lower()
        if direction is not None and direction not in [1, 0, -1]:
            raise ValueError("Direction must be in [1, 0, -1]")
        self.direction = 1 if direction is None else int(direction)  # eg planar norm
        self.basis = self._get_basis()

        self.fov_vert = 180 if fov_vert is None else fov_vert
        self.fov_horiz = 360 if fov_horiz is None else abs(fov_horiz)
        self.vert = False if vert is None else vert
        self.horiz = False if horiz is None else horiz

        # spacing and num points
        xr = abs(self.x2 - self.x1)
        yr = abs(self.y2 - self.y1)
        nx = min(max(int(xr * 10), 1), 50)
        ny = min(max(int(yr * 10), 1), 50)
        # default spacing
        default_xs = round(xr / nx, -int(np.floor(np.log10(xr / nx))))
        default_ys = round(yr / ny, -int(np.floor(np.log10(yr / ny))))
        # default number of points derived from default spacing
        default_num_x = int(round(xr / default_xs))
        default_num_y = int(round(yr / default_ys))

        if x_spacing is None:  # set from num_points if spacing is not provided
            self.num_x = default_num_x if num_x is None else int(num_x)
            self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, default_xs)
        else:
            self.x_spacing = x_spacing
            self.num_x = int(round(xr / self.x_spacing))
        if y_spacing is None:
            self.num_y = default_num_y if num_y is None else int(num_y)
            self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, default_ys)
        else:
            self.y_spacing = y_spacing
            self.num_y = int(round(yr / self.y_spacing))

        self._update()
        self.values = np.zeros(self.num_points)
        self.reflected_values = np.zeros(self.num_points)

    def _get_basis(self):
        """
        Return an orthonormal basis (u, v, n) for the surface:
        - n is the normal vector (points outward from surface)
        - u, v span the surface plane
        """

        if self.ref_surface == "xy":
            n = np.array([0, 0, 1])
        elif self.ref_surface == "xz":
            n = np.array([0, 1, 0])
        elif self.ref_surface == "yz":
            n = np.array([1, 0, 0])
        if self.direction != 0:
            n *= self.direction

        # Generate arbitrary vector not parallel to n
        tmp = np.array([1, 0, 0]) if abs(n[0]) < 0.9 else np.array([0, 1, 0])

        u = np.cross(n, tmp)
        u = u / np.linalg.norm(u)

        v = np.cross(n, u)
        basis = np.stack([u, v, n], axis=1)

        return basis

    def set_height(self, height):
        """set height of calculation plane. currently we only support vertical planes"""
        if type(height) not in [float, int]:
            raise TypeError("Height must be numeric")
        self.height = height
        self._update()
        return self

    def set_ref_surface(self, ref_surface):
        """
        set the reference surface of the calc plane--must be a string in [`xy`,`xz`,`yz`]
        """
        if not isinstance(ref_surface, str):
            raise TypeError("ref_surface must be a string in [`xy`,`xz`,`yz`]")
        if ref_surface.lower() not in ["xy", "xz", "yz"]:
            raise ValueError("ref_surface must be a string in [`xy`,`xz`,`yz`]")
        self.ref_surface = ref_surface
        self.basis = self._get_basis()
        self._update()
        return self

    def set_direction(self, direction):
        """
        set the direction of the plane normal
        Valid values are currently 1, -1 and 0
        """
        if direction not in [1, 0, -1]:
            raise ValueError("Direction must be in [1, 0, -1]")
        self.direction = int(direction)
        self.basis = self._get_basis()
        return self

    def set_dimensions(self, x1=None, x2=None, y1=None, y2=None, preserve_spacing=True):
        """set the dimensions and update the coordinate points"""
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self.dimensions = ((self.x1, self.y1), (self.x2, self.y2))
        # update number of points, keeping spacing
        xr = abs(self.x2 - self.x1)
        yr = abs(self.y2 - self.y1)
        if preserve_spacing:
            self.num_x = max(int(round(xr / self.x_spacing)), 1)
            self.num_y = max(int(round(yr / self.y_spacing)), 1)
        else:
            deg_x = -int(np.floor(np.log10(xr / self.num_x)))
            deg_y = -int(np.floor(np.log10(yr / self.num_y)))
            default_xs = round(xr / self.num_x, deg_x)
            default_ys = round(yr / self.num_y, deg_y)
            self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, default_xs)
            self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, default_ys)
        self._update()
        return self

    def set_spacing(self, x_spacing=None, y_spacing=None):
        """set the fineness of the grid spacing and update the coordinate points"""
        self.x_spacing = self.x_spacing if x_spacing is None else abs(x_spacing)
        self.y_spacing = self.y_spacing if y_spacing is None else abs(y_spacing)
        self.num_x = int(round(abs(self.x2 - self.x1) / self.x_spacing))
        self.num_y = int(round(abs(self.y2 - self.y1) / self.y_spacing))
        self._update()
        return self

    def set_num_points(self, num_x=None, num_y=None):
        """
        set the number of points desired in a dimension, instead of setting the spacing
        """
        self.num_x = self.num_x if num_x is None else abs(int(num_x))
        self.num_y = self.num_y if num_y is None else abs(int(num_y))
        if self.num_x == 0:
            warnings.warn("Number of x points must be at least 1")
            self.num_x += 1
        if self.num_y == 0:
            warnings.warn("Number of y points must be at least 1")
            self.num_y += 1

        # update spacing if required
        self.x_spacing = self._set_spacing(self.x1, self.x2, self.num_x, self.x_spacing)
        self.y_spacing = self._set_spacing(self.y1, self.y2, self.num_y, self.y_spacing)

        self._update()
        return self

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this calc zone must be recalculated
        """
        return [
            self.offset,
            self.x1,
            self.x2,
            self.x_spacing,
            self.y1,
            self.y2,
            self.y_spacing,
            self.height,
            self.ref_surface,
            self.direction,  # only for reflectance...possibly can be optimized
        ]

    def get_update_state(self):
        """
        return a set of parameters that, if changed, indicate that the
        calc zone need not be be recalculated, but may need updating
        """
        return [self.fov_vert, self.fov_horiz, self.vert, self.horiz]

    def _update(self):
        """
        Update the normal and number of points based on the spacing, and then the points
        """
        if self.x1 == self.x2:
            self.num_x = 1
        if self.y1 == self.y2:
            self.num_y = 1

        x_offset = min(self.x1, self.x2)
        y_offset = min(self.y1, self.y2)
        xp = np.array([i * self.x_spacing + x_offset for i in range(self.num_x)])
        yp = np.array([i * self.y_spacing + y_offset for i in range(self.num_y)])
        if self.offset:
            xp += (abs(self.x2 - self.x1) - abs(xp[-1] - xp[0])) / 2
            yp += (abs(self.y2 - self.y1) - abs(yp[-1] - yp[0])) / 2

        self.xp, self.yp = xp, yp
        self.points = [self.xp, self.yp]

        X, Y = [grid.reshape(-1) for grid in np.meshgrid(*self.points, indexing="ij")]

        if self.ref_surface in ["xy"]:
            Z = np.full(X.shape, self.height)
        elif self.ref_surface in ["xz"]:
            Z = Y
            Y = np.full(Y.shape, self.height)
        elif self.ref_surface in ["yz"]:
            Z = Y
            Y = X
            X = np.full(X.shape, self.height)

        self.coords = np.stack([X, Y, Z], axis=-1)
        self.num_points = np.array([len(self.xp), len(self.yp)])

    def plot_plane(self, fig=None, ax=None, vmin=None, vmax=None, title=None):
        """Plot the image of the radiation pattern"""
        if fig is None:
            if ax is None:
                fig, ax = plt.subplots()
            else:
                fig = plt.gcf()
        else:
            if ax is None:
                ax = fig.axes[0]

        title = "" if title is None else title
        values = self.get_values()
        if values is not None:
            vmin = values.min() if vmin is None else vmin
            vmax = values.max() if vmax is None else vmax
            extent = [self.x1, self.x2, self.y1, self.y2]

            values = values.T[::-1]
            img = ax.imshow(
                values, extent=extent, vmin=vmin, vmax=vmax, cmap=self.colormap
            )
            cbar = fig.colorbar(img, pad=0.03)
            ax.set_title(title)
            cbar.set_label(self.units, loc="center")
        return fig, ax

    def _write_rows(self):
        """
        export solution to csv
        """

        # if self.ref_surface == "xy":
        # xpoints = self.points[0].tolist()
        # ypoints = self.points[1].tolist()
        # elif self.ref_surface == "xz":
        # xpoints = self.points[0].tolist()
        # ypoints = [self.height] * self.num_y
        # elif self.ref_surface == "yz":
        # xpoints = [self.height] * self.num_x
        # ypoints = self.points[1].tolist()

        # num_x, num_y, *rest = [len(np.unique(val)) for val in self.coords.T if len(np.unique(val))>1]

        # points = [np.unique(val) for val in self.coords.T]
        # num_x, num_y, *rest = [len(val) for val in points if len(val) > 1]
        num_x, num_y = self.num_x, self.num_y  # tmp

        values = self.get_values()
        if values is None:
            vals = [[-1] * num_y] * num_x
        elif values.shape != (num_x, num_y):
            vals = [[-1] * num_y] * num_x
        else:
            vals = values
        zvals = self.coords.T[2].reshape(num_x, num_y).T[::-1]

        xpoints = self.coords.T[0].reshape(num_x, num_y).T[0].tolist()
        ypoints = self.coords.T[1].reshape(num_x, num_y)[0].tolist()

        if len(np.unique(xpoints)) == 1 and len(np.unique(ypoints)) == 1:
            xpoints = self.coords.T[0].reshape(num_x, num_y)[0].tolist()
            ypoints = self.coords.T[1].reshape(num_x, num_y).T[0].tolist()
            vals = vals.T
            zvals = zvals.T

        rows = [[""] + xpoints]

        rows += np.concatenate(([np.flip(ypoints)], vals)).T.tolist()
        rows += [""]
        # zvals

        rows += [[""] + list(line) for line in zvals]
        return rows



import numpy as np
import copy
from .calc_zone import CalcPlane


class ReflectanceManager:
    """
    Class for managing reflective surfaces and their interactions

    Attributes:
    x,y,z: floats
        cartesian dimensions--should be the same as the Room object
    reflectances: dict { str: float}
        dict with keys [`floor`, `ceiling`, `south`, `north`, `east`, `west`]
        and float values between 0 and 1. All values default 0.0
    x_spacings, y_spacings: dict { str: float}
        dict with same keys as `reflectances` and float values greater than 0.
        All values default None. Determines spacing in each reflective surface.
        Used preferentially over num_x/num_y if not None.
    num_x, num_y: dict { str: int}
        dict with same keys as `reflectances` and int values 1 or greater.
        All values default 10. Determines number of points in each reflective surface.
        Ignored if x_spacing/y_spacing is set
    max_num_passes: int, default=100
        When calculating interreflections, the maximum number of passes before
        the calculation concludes.
    threshold: float in [0,1], default=0.02
        When calculating interreflections, the threshold below which additional
        reflection contributions are no longer calculated. Interreflection
        calculation will step when the number of loops reaches max_num_passes
        or when the contributions fall below the threshold times the initial value,
        whichever happens first.
    """

    def __init__(
        self,
        x,
        y,
        z,
        reflectances=None,
        x_spacings=None,
        y_spacings=None,
        num_x=None,
        num_y=None,
        max_num_passes=None,
        threshold=None,
    ):

        self.x = x
        self.y = y
        self.z = z

        self.keys = ["floor", "ceiling", "south", "north", "east", "west"]
        default_reflectances = {surface: 0.0 for surface in self.keys}
        default_spacings = {surface: None for surface in self.keys}
        default_nums = {surface: 10 for surface in self.keys}

        self.reflectances = {**default_reflectances, **(reflectances or {})}
        self.x_spacings = {**default_spacings, **(x_spacings or {})}
        self.y_spacings = {**default_spacings, **(y_spacings or {})}
        self.num_x = {**default_nums, **(num_x or {})}
        self.num_y = {**default_nums, **(num_y or {})}
        self.max_num_passes = 100 if max_num_passes is None else int(max_num_passes)
        self.threshold = 0.02 if threshold is None else threshold
        if not isinstance(self.threshold, float):
            raise TypeError("threshold must be a float between 0 and 1")
        if self.threshold < 0 or self.threshold > 1:
            raise ValueError("threshold must be a float between 0 and 1")

        self.zone_dict = {}  # will contain all values from all contributions
        self.surfaces = {}
        self.managers = {}
        self._initialize_surfaces()

    def set_reflectance(self, R, wall_id=None):
        """set reflectance by wall_id or, if wall_if is None, to all walls"""
        if wall_id is None:
            # set this value for all walls
            for wall in self.keys:
                # self.reflectances[wall] = R
                self.surfaces[wall].R = R
        else:
            if wall_id not in self.keys:
                raise KeyError(f"wall_id must be in {self.keys}")
            else:
                # self.reflectances[wall_id] = R
                self.surfaces[wall_id].R = R
        # update the dict
        for key, val in self.surfaces.items():
            self.reflectances[key] = val.R

    def set_spacing(self, x_spacing=None, y_spacing=None, wall_id=None):
        """set x and y spacing by wall_id or, if wall_if is None, to all walls"""
        if wall_id is None:
            # set this value for all walls
            for wall in self.keys:
                self.surfaces[wall].plane.set_spacing(
                    x_spacing=x_spacing, y_spacing=y_spacing
                )
        else:
            if wall_id not in self.keys:
                raise KeyError(f"wall_id must be in {self.keys}")
            else:
                self.surfaces[wall_id].plane.set_spacing(
                    x_spacing=x_spacing, y_spacing=y_spacing
                )
        self._update_points()

    def set_num_points(self, num_x=None, num_y=None, wall_id=None):
        """set number of x and y points by wall_id or, if wall_if is None, to all walls"""
        if wall_id is None:
            for wall in self.keys:
                # set for all walls
                self.surfaces[wall].plane.set_num_points(num_x=num_x, num_y=num_y)
        else:
            if wall_id not in self.keys:
                raise KeyError(f"wall_id must be in {self.keys}")
            else:
                self.surfaces[wall_id].plane.set_num_points(num_x=num_x, num_y=num_y)
        self._update_points()

    def _initialize_surfaces(self):

        for wall, reflectance in self.reflectances.items():
            (
                x1,
                x2,
                y1,
                y2,
                height,
                ref_surface,
                direction,
            ) = self._get_surface_dimensions(wall)
            plane = CalcPlane(
                zone_id=wall,
                x1=x1,
                x2=x2,
                y1=y1,
                y2=y2,
                height=height,
                ref_surface=ref_surface,
                direction=direction,
                horiz=True,
                x_spacing=self.x_spacings[wall],
                y_spacing=self.y_spacings[wall],
                num_x=self.num_x[wall],
                num_y=self.num_y[wall],
            )
            self.surfaces[wall] = ReflectiveSurface(R=reflectance, plane=plane)
        self._update_points()
        # possibly this should go here? idk
        self.managers = {}
        for wall, surface in self.surfaces.items():
            ref_manager = copy.deepcopy(self)
            del ref_manager.surfaces[wall]
            self.managers[wall] = ref_manager

    def _update_points(self):
        for key, val in self.surfaces.items():
            self.x_spacings[key] = val.plane.x_spacing
            self.y_spacings[key] = val.plane.y_spacing
            self.num_x[key] = val.plane.num_x
            self.num_y[key] = val.plane.num_y

    def _get_surface_dimensions(self, wall):
        """retrieve the dimensions of a particular wall based on its id"""

        if wall == "floor":
            x1, x2, y1, y2 = 0, self.x, 0, self.y
            height = 0
            ref_surface = "xy"
            direction = 1
        elif wall == "ceiling":
            x1, x2, y1, y2 = 0, self.x, 0, self.y
            height = self.z
            ref_surface = "xy"
            direction = -1
        elif wall == "south":
            x1, x2, y1, y2 = 0, self.x, 0, self.z
            height = 0
            ref_surface = "xz"
            direction = 1
        elif wall == "north":
            x1, x2, y1, y2 = 0, self.x, 0, self.z
            height = self.y
            ref_surface = "xz"
            direction = -1
        elif wall == "west":
            x1, x2, y1, y2 = 0, self.y, 0, self.z
            height = 0
            ref_surface = "yz"
            direction = 1
        elif wall == "east":
            x1, x2, y1, y2 = 0, self.y, 0, self.z
            height = self.x
            ref_surface = "yz"
            direction = -1
        return x1, x2, y1, y2, height, ref_surface, direction

    def update_dimensions(self, x=None, y=None, z=None):
        """update the wall dimensions based on changes to the Room parent class"""
        self.x = self.x if x is None else x
        self.y = self.y if y is None else y
        self.z = self.z if z is None else z
        for wall, surface in self.surfaces.items():
            x1, x2, y1, y2, height, _, _ = self._get_surface_dimensions(wall)
            surface.plane.height = height
            surface.plane.set_dimensions(x1, x2, y1, y2, preserve_spacing=False)

    def calculate_incidence(self, lamps, hard=False):
        """
        calculate the incident irradiances on all reflective walls
        """

        # first pass
        for wall, surface in self.surfaces.items():
            surface.calculate_incidence(lamps, hard=hard)
        # subsequent passes
        self._interreflectance(lamps, hard=hard)

    def _interreflectance(self, lamps, hard=False):
        """
        calculate additional interreflectance
        """
        # create dict of ref managers for each wall
        managers = self._create_managers()

        # for storing the progressively increasing reflectance values
        dct = {}
        for wall, surface in self.surfaces.items():
            dct[wall] = surface.plane.values

        i = 0  # increases to self.max_num_passes
        percent = 1  # falls to self.threshold
        while percent > self.threshold and i < self.max_num_passes:
            pc = []
            for wall, surface in self.surfaces.items():

                init = surface.plane.values.mean()

                surface.calculate_incidence(
                    lamps, ref_manager=managers[wall], hard=hard
                )

                dct[wall] += surface.plane.reflected_values

                final = surface.plane.values.mean()
                if final > 0:
                    pc.append((abs(final - init) / final))
                else:
                    pc.append(0)
            percent = np.mean(pc)
            managers = self._update_managers(managers)
            i = i + 1
        for wall, surface in self.surfaces.items():
            surface.plane.values = dct[wall]

    def _update_managers(self, managers: dict) -> dict:
        """Update all interreflection managers with newly calculated surface incidences"""
        for wall, manager in managers.items():
            subwalls = list(manager.surfaces.keys())  # Create a static copy of keys
            for subwall in subwalls:
                # Update the values without modifying the dictionary structure
                np.copyto(
                    manager.surfaces[subwall].plane.values,
                    self.surfaces[subwall].plane.values,
                )
                # create new plane
                new_plane = copy.deepcopy(self.surfaces[subwall].plane)
                # replace the total values with only the reflected values
                new_plane.values = new_plane.reflected_values
                # Replace the object instead of deleting in-place
                manager.surfaces[subwall] = ReflectiveSurface(
                    R=self.surfaces[subwall].R, plane=new_plane
                )

        return managers  # Updated in place

    def _create_managers(self):
        """
        create a dict of reflection managers for each wall
        """
        managers = {}
        for wall, surface in self.surfaces.items():
            ref_manager = ReflectanceManager(
                x=self.x,
                y=self.y,
                z=self.z,
                reflectances=self.reflectances,
                x_spacings=self.x_spacings,
                y_spacings=self.y_spacings,
            )
            # assign planes
            for subwall, surface in ref_manager.surfaces.items():
                new_plane = copy.deepcopy(self.surfaces[subwall].plane)
                ref_manager.surfaces[subwall].plane = new_plane
            # remove the surface being reflected upon
            del ref_manager.surfaces[wall]
            managers[wall] = ref_manager
        return managers

    def calculate_reflectance(self, zone, hard=False):
        """
        calculate the reflectance contribution to a calc zone from each surface
        """

        threshold = zone.base_values.mean() * 0.01  # 1% of total value

        total_values = {}
        for wall, surface in self.surfaces.items():
            if surface.R * surface.plane.values.mean() > threshold:
                surface.calculate_reflectance(zone, hard=hard)
                values = surface.zone_dict[zone.zone_id]["values"]
                total_values[wall] = values
            else:
                total_values[wall] = np.zeros(zone.num_points)
        self.zone_dict[zone.zone_id] = total_values

        return total_values

    def get_total_reflectance(self, zone):
        """sum over all surfaces to get the total reflected values for that calc zone"""
        dct = self.zone_dict[zone.zone_id]
        values = np.zeros(zone.num_points).astype("float32")
        for wall, surface_vals in dct.items():
            if surface_vals is not None:
                values += surface_vals * self.reflectances[wall]
        return values


class ReflectiveSurface:
    """
    Class that represents a single reflective surface defined by a calculation
    zone and a float value R between 0 and 1.
    """

    def __init__(self, R, plane):

        if not isinstance(R, (float, int)):
            raise TypeError("R must be a float in range [0, 1]")
        if R > 1 or R < 0:
            raise ValueError("R must be a float in range [0, 1]")

        if not isinstance(plane, CalcPlane):
            raise TypeError("plane must be a CalcPlane object")

        self.R = R
        self.plane = plane
        self.num_passes = 0  # init
        self.zone_dict = {}

    def calculate_incidence(self, lamps, ref_manager=None, hard=False):
        """calculate incoming radiation onto all surfaces"""
        self.plane.calculate_values(lamps=lamps, ref_manager=ref_manager, hard=hard)

    def get_calc_state(self):
        """check if the surface needs to be recalculated"""
        return self.plane.get_calc_state()

    def get_update_state(self):
        """"""
        return self.plane.get_update_state() + [self.plane.values.sum()]

    def calculate_reflectance(self, zone, hard=False):
        """
        calculate the reflective contribution of this reflective surface
        to a provided calculation zone

        Arguments:
            zone: a calculation zone onto which reflectance is calculated
            lamp: optional. if provided, and incidence not yet calculated, uses this
            lamp to calculate incidence. mostly this is just for
        """
        # first make sure incident irradiance is calculated
        if self.plane.values is None:
            raise ValueError("Incidence must be calculated before reflectance")

        if self.zone_dict.get(zone.zone_id) is None:
            self.zone_dict[zone.zone_id] = {}

        calc_state = zone.get_calc_state()
        update_state = zone.get_update_state()
        surface_calc_state = self.get_calc_state()
        surface_update_state = self.get_update_state()
        NEW_ZONE = self.zone_dict[zone.zone_id].get("values") is None
        ZONE_RECALC = calc_state != self.zone_dict[zone.zone_id].get("calc_state")
        ZONE_UPDATE = update_state != self.zone_dict[zone.zone_id].get("update_state")
        SURF_RECALC = surface_calc_state != self.zone_dict[zone.zone_id].get(
            "surface_calc_state"
        )
        SURF_UPDATE = surface_update_state != self.zone_dict[zone.zone_id].get(
            "surface_update_state"
        )

        RECALCULATE = NEW_ZONE or ZONE_RECALC or SURF_RECALC or hard
        UPDATE = RECALCULATE or NEW_ZONE or ZONE_UPDATE or SURF_UPDATE or hard

        if RECALCULATE:
            form_factors, theta_zone = self._calculate_coordinates(zone)
        else:
            form_factors = self.zone_dict[zone.zone_id]["form_factors"]
            theta_zone = self.zone_dict[zone.zone_id]["theta_zone"]
        if UPDATE:
            I_r = self.plane.values[:, :, np.newaxis, np.newaxis, np.newaxis]
            element_size = self.plane.x_spacing * self.plane.y_spacing

            values = (I_r * element_size * form_factors).astype("float32")

            values = self._apply_filters(values, theta_zone, zone)

            # Sum over all self.plane points to get total values at each volume point
            values = np.sum(values, axis=(0, 1))  # Collapse the dimensions
            values = values.reshape(*zone.num_points)
        else:
            values = self.zone_dict[zone.zone_id]["values"]

        # update the state
        self.zone_dict[zone.zone_id] = {
            "form_factors": form_factors,
            "theta_zone": theta_zone,
            "values": values,
            "calc_state": calc_state,
            "update_state": update_state,
            "surface_calc_state": surface_calc_state,
            "surface_update_state": surface_update_state,
        }

        return (values * self.R).astype("float32")

    def _apply_filters(self, values, theta_zone, zone):
        """apply field-of-view based calculations"""

        # clean nans
        if np.isnan(values).any():
            values = np.ma.masked_invalid(values)

        if zone.calctype == "Plane":

            # apply normals
            if zone.direction != 0:
                values[theta_zone > np.pi / 2] = 0

            # apply vertical field of view
            values[theta_zone < (np.pi / 2 - np.radians(zone.fov_vert / 2))] = 0
            values[theta_zone > (np.pi / 2 + np.radians(zone.fov_vert / 2))] = 0

            if zone.vert:
                values *= np.sin(theta_zone)
            if zone.horiz:
                values *= abs(np.cos(theta_zone))

        return values

    def _calculate_coordinates(self, zone):
        """
        return the angles and distances between the points of the reflective
        surface and the calculation zone

        this is the expensive step!
        """
        surface_points = self.plane.coords.reshape(*self.plane.num_points, 3)
        zone_points = zone.coords.reshape(*zone.num_points, 3)

        differences = (
            surface_points[:, :, np.newaxis, np.newaxis, np.newaxis, :] - zone_points
        )
        x, y, z = differences.reshape(-1, 3).T
        distances = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        distances = distances.reshape(differences.shape[0:-1])
        # distances = np.linalg.norm(differences, axis=-1) # notably slower!

        # angles relative to reflective surface -- always between 0 and 90 unless the calc zone has been misspecified
        rel_surface = differences @ self.plane.basis
        cos_theta_surface = -rel_surface[..., 2] / distances
        cos_theta_surface[cos_theta_surface < 0] = 0
        # theta_surface = np.arccos(cos_theta_surface)
        form_factors = cos_theta_surface / (np.pi * distances ** 2)
        form_factors = form_factors.astype("float32")

        #  angles relative to calculation zone. only relevant for planes
        if zone.calctype == "Plane":
            rel_zone = differences @ zone.basis
            cos_theta_zone = rel_zone[..., 2] / distances
            theta_zone = np.arccos(cos_theta_zone).astype("float32")
        else:
            theta_zone = None

        # # ? absolute? angles
        # cos_theta = -differences[..., 2] / distances
        # theta = np.arccos(cos_theta)

        return form_factors, theta_zone



import warnings
import pathlib
from pathlib import Path
import datetime
import json
import zipfile
import io
import csv
import plotly.io as pio
from plotly.graph_objs._figure import Figure as plotly_fig
from matplotlib.figure import Figure as mpl_fig

# -------------- Loading a room from file -------------------


def load_room(filedata):
    """
    TODO: wow this is hideous this should really just be a class method
    load a room object from json filedata
    """
    from .room import Room

    load_data = _parse_loadfile(filedata)
    saved_version = load_data["guv-calcs_version"]
    current_version = get_version(Path(__file__).parent / "_version.py")
    if saved_version != current_version:
        warnings.warn(
            f"This file was saved with guv-calcs {saved_version}, while you have {current_version} installed."
        )
    room_dict = load_data["data"]
    return Room.from_dict(room_dict)


def _parse_loadfile(filedata):
    """
    validate and parse a loadfile
    """

    if isinstance(filedata, (str, bytes or bytearray)):
        try:
            dct = json.loads(filedata)
        except json.JSONDecodeError:
            path = Path(filedata)
            dct = _load_file(path)
    elif isinstance(filedata, pathlib.PurePath):
        dct = _load_file(filedata)

    return dct


def _load_file(path):
    """load json from a path"""
    if path.is_file():
        if path.suffix.lower() != ".guv":
            raise ValueError("Please provide a valid .guv file")
        with open(path, "r") as json_file:
            try:
                dct = json.load(json_file)
            except json.JSONDecodeError:
                raise ValueError(".guv file is malformed")
    else:
        raise FileNotFoundError("Please provide a valid .guv file")
    return dct


def save_room(room, fname):
    """save all relevant parameters to a json file"""
    savedata = {}
    version = get_version(Path(__file__).parent / "_version.py")
    savedata["guv-calcs_version"] = version

    now = datetime.datetime.now()
    now_local = datetime.datetime.now(now.astimezone().tzinfo)
    timestamp = now_local.strftime("%Y-%m-%d %H:%M:%S %Z%z")
    savedata["timestamp"] = timestamp

    savedata["data"] = room.to_dict()
    if fname is not None:
        filename = _check_savefile(fname, ".guv")
        with open(filename, "w") as json_file:
            json.dump(savedata, json_file, indent=4)
    else:
        return json.dumps(savedata, indent=4)


def _check_savefile(filename, ext):
    """
    enforce that a savefile has the correct extension
    """

    if not ext.startswith("."):
        ext = "." + ext

    if isinstance(filename, str):
        if not filename.lower().endswith(ext):
            filename += ext
    elif isinstance(filename, pathlib.PurePath):
        if not filename.suffix == ext:
            filename = filename.parent / (filename.name + ext)
    return filename


def export_room_zip(
    room,
    fname=None,
    include_plots=False,
    include_lamp_files=False,
    include_lamp_plots=False,
):

    """
    write the room project file and all results files to a zip file. Optionally include
    extra files like lamp ies files, spectra files, and plots.
    """

    # save base project
    data_dict = {"room.guv": room.save()}

    # save all results
    for zone_id, zone in room.scene.calc_zones.items():
        if zone.calctype != "Zone":
            data_dict[zone.name + ".csv"] = zone.export()
            if include_plots:
                if zone.dose:
                    title = f"{zone.hours} Hour Dose"
                else:
                    title = "Irradiance"
                if zone.calctype == "Plane":
                    # Save the figure to a BytesIO object
                    title += f" ({zone.height} m)"
                    fig, ax = zone.plot_plane(title=title)
                    data_dict[zone.name + ".png"] = fig_to_bytes(fig)
                elif zone.calctype == "Volume":
                    fig = zone.plot_volume()
                    data_dict[zone.name + ".png"] = fig_to_bytes(fig)

    # save lamp files if indicated to
    for lamp_id, lamp in room.scene.lamps.items():
        if lamp.filedata is not None:
            if include_lamp_files:
                data_dict[lamp.name + ".ies"] = lamp.save_ies()
            if include_lamp_plots:
                ies_fig, ax = lamp.plot_ies(title=lamp.name)
                data_dict[lamp.name + "_ies.png"] = fig_to_bytes(ies_fig)
        if lamp.spectra is not None:
            if include_lamp_plots:
                linfig, _ = lamp.spectra.plot(
                    title=lamp.name, yscale="linear", weights=True, label=True
                )
                logfig, _ = lamp.spectra.plot(
                    title=lamp.name, yscale="log", weights=True, label=True
                )
                linkey = lamp.name + "_spectra_linear.png"
                logkey = lamp.name + "_spectra_log.png"
                data_dict[linkey] = fig_to_bytes(linfig)
                data_dict[logkey] = fig_to_bytes(logfig)

    zip_buffer = io.BytesIO()
    # Create a zip file within this BytesIO object
    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        # Loop through the dictionary, adding each string/byte stream to the zip
        for filename, content in data_dict.items():
            # Ensure the content is in bytes
            if isinstance(content, str):
                content = content.encode("utf-8")
            # Add the file to the zip; writing the bytes to a BytesIO object for the file
            file_buffer = io.BytesIO(content)
            zip_file.writestr(filename, file_buffer.getvalue())
    zip_bytes = zip_buffer.getvalue()

    if fname is not None:
        with open(fname, "wb") as f:
            f.write(zip_bytes)
    else:
        return zip_bytes


def generate_report(self, fname=None):
    """
    Dump a one file CSV with a snapshot of the current room.
    Sections are separated by blank lines so Excel shows each
    header group clearly.
    """
    precision = self.precision if self.precision > 3 else 3
    fmt = lambda v: round(v, precision) if isinstance(v, (int, float)) else v
    # ───  Room parameters  ───────────────────────────────
    rows = [["Room Parameters"]]
    rows += [["", "Dimensions", "x", "y", "z", "units"]]
    d = self.dim
    rows += [["", "", fmt(d.x), fmt(d.y), fmt(d.z), d.units]]
    vol_units = "ft 3" if self.units == "feet" else "m 3"
    rows += [["", "Volume", fmt(self.volume), vol_units]]
    rows += [[""]]

    # ───  Reflectance  ──────────────────────────────────
    rows += [["", "Reflectance"]]
    rows += [["", "", "Floor", "Ceiling", "North", "South", "East", "West", "Enabled"]]
    rows += [["", "", *self.ref_manager.reflectances.values(), self.enable_reflectance]]
    rows += [[""]]

    # ───  Luminaires  ───────────────────────────────────
    if self.scene.lamps:
        rows += [["Luminaires"]]
        rows += [["", "", "", "Surface Position", "", "", "Aim"]]
        rows += [
            [
                "",
                "ID",
                "Name",
                "x",
                "y",
                "z",
                "x",
                "y",
                "z",
                "Orientation",
                "Tilt",
                "Surface Length",
                "Surface Width",
                "Fixture Depth",
            ]
        ]
        for lamp in self.scene.lamps.values():
            rows += [
                [
                    "",
                    lamp.lamp_id,
                    lamp.name,
                    fmt(lamp.x),
                    fmt(lamp.y),
                    fmt(lamp.z),
                    fmt(lamp.aimx),
                    fmt(lamp.aimy),
                    fmt(lamp.aimz),
                    fmt(lamp.heading),
                    fmt(lamp.bank),
                    fmt(lamp.surface.length),
                    fmt(lamp.surface.width),
                    fmt(lamp.surface.depth),
                ]
            ]
        rows += [[""]]

    # ───  Calculation planes  ───────────────────────────
    planes = [z for z in self.scene.calc_zones.values() if z.calctype == "Plane"]
    if planes:
        rows += [["Calculation Planes"]]
        rows += [
            [
                "",
                "ID",
                "Name",
                "x1",
                "x2",
                "y1",
                "y2",
                "height",
                "Vertical irradiance",
                "Horizontal irradiance",
                "Vertical field of view",
                "Horizontal field of view",
                "Dose",
                "Dose Hours",
            ]
        ]
        for pl in planes:
            rows += [
                [
                    "",
                    pl.zone_id,
                    pl.name,
                    fmt(pl.x1),
                    fmt(pl.x2),
                    fmt(pl.y1),
                    fmt(pl.y2),
                    fmt(pl.height),
                    pl.vert,
                    pl.horiz,
                    pl.fov_vert,
                    pl.fov_horiz,
                    pl.dose,
                    pl.hours if pl.dose else "",
                ]
            ]
        rows += [[""]]

    # ───  Calculation volumes  ──────────────────────────
    vols = [z for z in self.scene.calc_zones.values() if z.calctype == "Volume"]
    if vols:
        rows += [["Calculation Volumes"]]
        rows += [
            [
                "",
                "ID",
                "Name",
                "x1",
                "x2",
                "y1",
                "y2",
                "z1",
                "z2",
                "Dose",
                "Dose Hours",
            ]
        ]
        for v in vols:
            rows += [
                [
                    "",
                    v.zone_id,
                    v.name,
                    fmt(v.x1),
                    fmt(v.x2),
                    fmt(v.y1),
                    fmt(v.y2),
                    fmt(v.z1),
                    fmt(v.z2),
                    v.dose,
                    v.hours if v.dose else "",
                ]
            ]
        rows += [[""]]

    # ───  Statistics  ──────────────────────────
    zones = [z for z in self.scene.calc_zones.values() if z.calctype != "Zone"]
    if zones:
        rows += [["Statistics"]]
        rows += [
            ["", "Calculation Zone", "Avg", "Max", "Min", "Max/Min", "Avg/Min", "Units"]
        ]
        for zone in zones:
            values = zone.get_values()
            avg = values.mean()
            mx = values.max()
            mn = values.min()
            mxmin = mx / mn
            avgmin = avg / mn
            rows += [
                [
                    "",
                    zone.name,
                    round(avg, precision),
                    round(mx, precision),
                    round(mn, precision),
                    round(mxmin, precision),
                    round(avgmin, precision),
                    zone.units,
                ]
            ]
        rows += [[""]]

    # footer
    rows += [[f"Generated {datetime.datetime.now().isoformat(timespec='seconds')}"]]
    csv_bytes = rows_to_bytes(rows)

    if fname is not None:
        with open(fname, "wb") as csvfile:
            csvfile.write(csv_bytes)
    else:
        return csv_bytes


# ------- Conversions to bytes ---------------


def fig_to_bytes(fig):
    if isinstance(fig, mpl_fig):
        buf = io.BytesIO()
        fig.savefig(
            buf, format="png"
        )  # You can change the format as needed (e.g., 'jpeg', 'pdf')
        buf.seek(0)  # Rewind the buffer
        byt = buf.getvalue()
    elif isinstance(fig, plotly_fig):
        byt = pio.to_image(fig, format="png", scale=1)
    else:
        raise TypeError("This figure type cannot be converted to bytes")
    return byt


def rows_to_bytes(rows, encoding="cp1252"):
    buffer = io.StringIO()
    writer = csv.writer(buffer)
    writer.writerows(rows)

    # Get the CSV data from buffer, convert to bytes
    csv_data = buffer.getvalue()
    csv_bytes = csv_data.encode(encoding)  # encode to bytes
    return csv_bytes


# ----------- misc io ----------


def load_csv(datasource):
    """load csv data from either path or bytes"""
    if isinstance(datasource, (str, pathlib.PurePath)):
        filepath = Path(datasource)
        filetype = filepath.suffix.lower()
        if filetype != ".csv":
            raise TypeError("Currently, only .csv files are supported")
        csv_data = open(datasource, mode="r")
    elif isinstance(datasource, bytes):
        # Convert bytes to a string using io.StringIO to simulate a file
        csv_data = io.StringIO(datasource.decode("utf-8"), newline="")
    else:
        raise TypeError(f"File type {type(datasource)} not valid")
    return csv_data


def get_version(path) -> dict:

    version = {}
    with open(path) as f:
        exec(f.read(), version)
    return version["__version__"]



from dataclasses import dataclass, replace
import numpy as np


@dataclass(frozen=True, slots=True)
class RoomDimensions:
    x: float
    y: float
    z: float
    units: str = "meters"

    def volume(self) -> float:
        return self.x * self.y * self.z

    def dimensions(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])

    def with_(self, *, x=None, y=None, z=None, units=None):
        return replace(
            self,
            x=self.x if x is None else x,
            y=self.y if y is None else y,
            z=self.z if z is None else z,
            units=self.units if units is None else units.lower(),
        )



AVAILABLE_UNITS = ["meters", "feet"]


def meters_to_feet(arg):
    return arg / 0.3048


def feet_to_meters(arg):
    return arg * 0.3048


def convert(function, *args):
    result = tuple(function(arg) if arg is not None else None for arg in args)
    return result[0] if len(result) == 1 else result


def convert_units(src, dst, *args):
    """general purpose function for converting between units"""
    if not (isinstance(src, str) and isinstance(dst, str)):
        raise ValueError("Unit labels must be strings")

    src, dst = src.lower(), dst.lower()

    if src not in AVAILABLE_UNITS:
        raise ValueError(f"{src} is not an available unit")
    if dst not in AVAILABLE_UNITS:
        raise ValueError(f"{dst} is not an available unit")

    if src == dst:
        result = args
        if len(result) == 1:
            result = result[0]
    elif src == "meters":
        if dst == "feet":
            result = convert(meters_to_feet, *args)
    elif src == "feet":
        if dst == "meters":
            result = convert(feet_to_meters, *args)
    else:
        raise ValueError("Something went wrong that really should not have gone wrong")

    return result
